//----------------------------------------------------------------------------
// Controller for Magnetic Levitation Kit
// Department of Mechanical Engineering, Numazu College of Technology
// 
// Project: maglev_a.soc
// Version: 1.00
//
// Sampling time: 2 msec
//
// Mode
// 0:Manual Adjust
// 1:Auto Adjust
// 2:External Control(Slave Mode)
//-----------------------------------------------------------------------------

#include <m8c.h>                         // part specific constants and macros
#include "PSoCAPI.h"                     // PSoC API definitions for all User Modules

// Flash ROM Initialize
#pragma abs_address 0x3C00
     const signed char r_rom[8] ={0,0,0,0,0,0,0,0};
#pragma end_abs_address

// Sampling frequency
#define SAMPLING_FREQ 500

// Sensitivity setting of manual adjust
#define GAIN_SC 100
#define DGAIN_SC 100

// A/D Converter
#define ADC_RESOLUTION 0x01FF
#define REF_VOLTAGE 5

// PWM
#define PWM_RESOLUTION 1023
#define PWM_LIMIT 256 // Software Limitation (25%)

// Non-object sensor threshold
#define HALL_LIMIT 0x0008

// PGA
#define PGA_GAIN 4

// DAC
#define DAC_MAX 0x01FF

// Button check
#define BUTTON_WAIT 50

// Mode change
#define MODECHANGEWAIT 500

// Pin-assign for Buttons
#define BTNPORT PRT2DR
#define PINPLUSMINUS	0b01000000
#define PINRIGHT		0b00010000
#define PINENTER		0b00000100
#define PINLEFT			0b00000001
#define PINBITHIGHLOW	0b10000000

// Global variables
BYTE bPort1, bLastPort1;
char cTimerFlag;
char cBitHighLowFlag;
char cCountFlag;
char cStepNum;
char cLCDFlag;
char cCursor;
char cModeFlag;
int iAdValue;
int iAdValue1, iAdValue2, iAdValue3;
int iPwmValue;
int iPlusMinusFlag;
int iBufDGain, iBufPGain, iBufBias, iBufRef;
int iTmp;
int iLCDClearCount;
int iModeChangeCount;
int iE2promCount;
int iDef;
int iDefOld;
int iPValue,iDValue;
int iSum;
unsigned char ucBuf[8];
int iTmpT1, iTmpT2, iTmpT3, iTmpT4;
int iAdAverage;
int iAdVariance;
int iAdVariance2;
int iAACount;

// Timer interrupt handler
#pragma interrupt_handler INT_Timer8_1
void INT_Timer8_1(void) {
	cTimerFlag = 1;
}


// Main function
void main()
{
	int ii;
	/////////////////////////////
	// Initialize User Modules //
	/////////////////////////////
	
	// PGA_1
    PGA_1_Start(PGA_1_HIGHPOWER);
	// PGA_2
    PGA_2_Start(PGA_2_HIGHPOWER);
	// PGA_3
    PGA_3_Start(PGA_3_HIGHPOWER);
    // TRIADC_1
	TRIADC_1_Start(TRIADC_1_HIGHPOWER);
	TRIADC_1_GetSamples(0);
	// PWM16_1
	PWM16_1_Start();
	// Timer8_1
	Timer8_1_EnableInt();
	Timer8_1_Start();
	// DAC9_1
	DAC9_1_Start(DAC9_1_FULLPOWER);
	// DAC9_2
	DAC9_2_Start(DAC9_2_FULLPOWER);
	// LCD_1
    LCD_1_Start(); 
 	// Enable Global Interrupts
   	M8C_EnableGInt;

	///////////////////////////
	// Initialize Parameters //
	///////////////////////////
	iAdValue = 0;
	cTimerFlag = 0;
	iPlusMinusFlag = 0x0001;
	cBitHighLowFlag = 0x01;
	iBufDGain = 0;
	iBufPGain = 0;
	iBufBias = 0;
	iBufRef = 0;
	cCursor = 0;
	cCountFlag = 0;
	bLastPort1 = 0;
    cLCDFlag = 0;
    cModeFlag = 0;
    iLCDClearCount = 0;
    iModeChangeCount = 0;
    
	/////////////////////////
	// Main Loop (Endless) //
	/////////////////////////
    while (1)
    {
    	//####
		//## Timer Wait
		//#
		while(cTimerFlag == 0) ;
		cTimerFlag = 0;
		PRT0DR = PRT0DR ^ 0b01000000;

		//###
		//## Get Button State
		//#
		bPort1 = BTNPORT;
     
		//###
		//## A/D CNV Sampling
		//#
        while (TRIADC_1_fIsDataAvailable() == 0) ;
	    iAdValue1 = TRIADC_1_iGetData1(); // Sensor output(+)
	    iAdValue2 = TRIADC_1_iGetData2(); // External input
	    iAdValue3 = TRIADC_1_iGetData3(); // Sensor output(-)
	    TRIADC_1_ClearFlag();
		// Hall Voltage
        iAdValue = iAdValue1 - iAdValue3;
        
        //###
        //## Vout_hall (Sensor Output)
        //#
        iTmp = iAdValue;
        if(iTmp > DAC_MAX) iTmp = DAC_MAX;
        DAC9_1_WriteStall(iTmp);
		  
        //###
        //## Button Check
        //#
        if((cCountFlag >= BUTTON_WAIT)&&(cModeFlag < 2)) {	        
	        // Button Enter
	        if(bPort1 & PINENTER) {
	        	switch(cCursor) {
	        	case 0:
		        	iBufBias += cBitHighLowFlag * iPlusMinusFlag;
		        	if(iBufBias<0) iBufBias=0;
		        	if(iBufBias>0xfff) iBufBias=0xfff;
		        	break;
				case 1:
		        	iBufRef += (0x00FF & (int)cBitHighLowFlag) * iPlusMinusFlag;
		        	if(iBufRef<0) iBufRef=0;
		        	if(iBufRef>0xfff) iBufRef=0xfff;
		        	break;
				case 2:
		        	iBufPGain += (0x00FF & (int)cBitHighLowFlag) * iPlusMinusFlag;
		        	if(iBufPGain<0) iBufPGain=0;
		        	if(iBufPGain>0xfff) iBufPGain=0xfff;
		        	break;
				case 3:
		        	iBufDGain += (0x00FF & cBitHighLowFlag) * iPlusMinusFlag;
		        	if(iBufDGain<0) iBufDGain=0;
		        	if(iBufDGain>0xfff) iBufDGain=0xfff;
		       		break;
		       	default:
		       		break;
		       	}
	        }
	        cCountFlag = 0;
	    }
	    if(cCountFlag!=0xFF) cCountFlag++;
	    
        // Button Right
        if(bPort1 & ~bLastPort1 & PINRIGHT) {
        	if(++cCursor==0x04) cCursor = 3;
        }
        
        // Button Left
        if(bPort1 & ~bLastPort1 & PINLEFT) {
        	if(--cCursor==0xFF) cCursor = 0;
        }
        
        // Button PlusMinus
        if(bPort1 & ~bLastPort1 & PINPLUSMINUS) {
        	iPlusMinusFlag *= -1;
        }
        
        // Button BitHighLow
        if(bPort1 & ~bLastPort1 & PINBITHIGHLOW) {
        	cBitHighLowFlag ^= 0x11;
        }

        // LCD Reset
        if((bPort1 & PINRIGHT)&&(bPort1 & PINLEFT)) iLCDClearCount++;
        else iLCDClearCount=0;
        if(iLCDClearCount==MODECHANGEWAIT) LCD_1_Start();

        // Mode Change
        if((bPort1 & PINPLUSMINUS)&&(bPort1 & PINBITHIGHLOW)) iModeChangeCount++;
        else iModeChangeCount = 0;
        if(iModeChangeCount == MODECHANGEWAIT) {
            if(++cModeFlag == 3) cModeFlag = 0;
            cLCDFlag = 0;
            LCD_1_Start();
            if(cModeFlag == 1) {
	            cStepNum = 0;
				iBufDGain = 0;
				iBufPGain = 0;
				iBufBias = 0;
				iBufRef = 0;
				iAACount = 0;
				LCD_1_PrCString("Auto Adjust Mode");
				while(BTNPORT&PINPLUSMINUS);
			}
        }
        
        // EEPROM Write/Read
        if(bPort1 & PINPLUSMINUS) iE2promCount++;
        if(bPort1 & PINBITHIGHLOW) iE2promCount--;
        if((bPort1 & PINPLUSMINUS)+(bPort1 & PINBITHIGHLOW)==0) iE2promCount=0;
        if(iE2promCount==MODECHANGEWAIT) {
            ucBuf[0]=0x00FF&iBufBias;
            ucBuf[1]=(0xFF00&iBufBias)>>8;
            ucBuf[2]=0x00FF&iBufRef;
            ucBuf[3]=(0xFF00&iBufRef)>>8;
            ucBuf[4]=0x00FF&iBufPGain;
            ucBuf[5]=(0xFF00&iBufPGain)>>8;
            ucBuf[6]=0x00FF&iBufDGain;
            ucBuf[7]=(0xFF00&iBufDGain)>>8;
            iTmp = E2PROM_1_bE2Write(0,ucBuf,64,25);
            if(iTmp==0) {
	            LCD_1_Start();
    	        LCD_1_PrCString("Parameter Saved");
        	    LCD_1_Position(1,0);
            	LCD_1_PrHexInt(iBufBias);
            	LCD_1_PrHexInt(iBufRef);
            	LCD_1_PrHexInt(iBufPGain);
            	LCD_1_PrHexInt(iBufDGain);
            }
            else {
            	LCD_1_Start();
            	LCD_1_PrCString("Error");
            	LCD_1_PrHexInt(iTmp);
            }
            while(BTNPORT&PINPLUSMINUS);
            LCD_1_Start();
        }else if(iE2promCount==-MODECHANGEWAIT) {
            E2PROM_1_E2Read(0,ucBuf,8);
            iBufBias=ucBuf[0]+(ucBuf[1]<<8);
            iBufRef=ucBuf[2]+(ucBuf[3]<<8);
            iBufPGain=ucBuf[4]+(ucBuf[5]<<8);
            iBufDGain=ucBuf[6]+(ucBuf[7]<<8);
            LCD_1_Start();
            LCD_1_PrCString("Parameter Loaded");
            LCD_1_Position(1,0);
            LCD_1_PrHexInt(iBufBias);
            LCD_1_PrHexInt(iBufRef);
            LCD_1_PrHexInt(iBufPGain);
            LCD_1_PrHexInt(iBufDGain);
            while(BTNPORT&PINBITHIGHLOW);
            LCD_1_Start();
        }        
        bLastPort1 = bPort1;

		//###
		//## Auto Adjust Step
		//#
        if((cModeFlag==1)&&(cStepNum<11)) {
        	switch(cStepNum) {
        	case 0:
        		// 0.5sec wait
        		if(iAACount++ > (SAMPLING_FREQ >> 1)) {
        			cStepNum++;
        			iAACount = 0;
        		}
        		break;
        	case 1:
        	    iTmpT1 = iAdValue; // iTmpT1:Bottom sensor output
        	    cStepNum++;
        		break;
        	case 2:
        		iBufBias++; // Bias lamp up
        		// threshold:0x0010 
        		if(iAdValue > (iTmpT1 + 0x0010)) {
        			iTmpT2 = iBufBias; // TmpT2:Bottom pop PWM
        			cStepNum++;
        		}
        		break;
        	case 3:
        		// 0.5sec wait
        		if(iAACount++ > (SAMPLING_FREQ >> 1)) {
        			cStepNum++;
        			iAACount = 0;
        		}
        		break;
        	case 4:
        		iTmpT3 = iAdValue; // iTmpT3:Top sensor output
        		cStepNum++;
        		break;
        	case 5:
        		iBufBias--; // Bias lamp down
        		// threshold:0x0010 
        		if(iAdValue < (iTmpT3 - 0x0010) ) {
        			iTmpT4 = iBufBias; // iTmpT4:Top fall PWM
        			cStepNum++;
        		}
        		break;
        	case 6:
        		// 0.5sec wait
        		if(iAACount++ > (SAMPLING_FREQ >> 1)) {
        			cStepNum++;
        			iAACount = 0;
        		}
        		break;
        	case 7:
        		// Set Bias and Ref on half point
        		iBufBias = (iTmpT4 + iTmpT2)>>1;
        		iBufRef = (iTmpT3 + iTmpT1)>>1;
        		cStepNum++;
        		break;
        	case 8:
        		// 0.5sec wait
        		if(iAACount++ > (SAMPLING_FREQ >> 1)) {
        			cStepNum++;
        			iAACount = 0;
        		}
				break;
        	case 9:
        		iBufPGain++; // PGain lamp up
        		// threshold:0x0020
        		if(iAdValue > (iTmpT1 + 0x0020)) {
        			iAdAverage = iAdValue;
        			iAdVariance = 0x0400;
        			cStepNum++;
        		}
        		break;
        	case 10:
        		iBufDGain++; // DGain lamp up
   				// Evaluation Value: iAdVariance
        		iAdAverage = (7*iAdAverage + iAdValue)>>3;
        		iTmp = iAdValue - iAdAverage;
        		iAdVariance = (7*iAdVariance + iTmp*iTmp)>>3;
        		// threshold:0x0010
        		if(iAdVariance < 0x0010) cStepNum++;
        	    break;
        	default:
        		break;
        	}
        }

        //###
        //## PD Controller
        //#
        // A/D Value -> Input Voltage
        if(cModeFlag < 2) {
        	iDef = (iBufRef - iAdValue);

			// P Controller
			iPValue = iDef * iBufPGain;
			
			// D Controller
			iDValue = (iDef - iDefOld) * iBufDGain * (SAMPLING_FREQ / DGAIN_SC);
			iDefOld = iDef;
	 		
	 		// P + D
	 		iSum = (iPValue + iDValue) / GAIN_SC;
	
			// Set PWM Value
			iPwmValue = iSum * (PWM_RESOLUTION / ADC_RESOLUTION) / PGA_GAIN + iBufBias;
		}else{
		    iPwmValue = iAdValue2 * PWM_RESOLUTION / ADC_RESOLUTION; // 5V:100%
		}
		if(iPwmValue < 0) iPwmValue = 0;
		else if(iPwmValue > PWM_LIMIT) iPwmValue = PWM_LIMIT; //limit

		
		// Non-object PWM off
		if(iAdValue < HALL_LIMIT) iPwmValue = 0;

		// PWM duty ratio rewrite
 		PWM16_1_WritePulseWidth( iPwmValue );

        //###
        //## Vout_pwm (PWM duty ratio output)
        //#
 		iTmp = iPwmValue * ADC_RESOLUTION / PWM_RESOLUTION; // 100%:5V
 		if(iTmp > DAC_MAX) iTmp = DAC_MAX;
 		DAC9_2_WriteStall(iTmp);
 		        
        //###
        //## Display LCD
        //#
        if(cModeFlag < 2) {
	        switch(cLCDFlag) {
	        case 0:
		        LCD_1_Position(0,0);
		        LCD_1_PrCString("B");
		        LCD_1_PrHexByte((unsigned char)(iBufBias>>4));
		        break;
		    case 1:
		        LCD_1_PrHexByte((unsigned char)((iBufBias<<4)&0x00FF));
		        LCD_1_Position(0,4);
		        LCD_1_PrCString("R");
		        break;
		    case 2:
		        LCD_1_PrHexByte((unsigned char)(iBufRef>>4));
		        break;
		    case 3:
		        LCD_1_PrHexByte((unsigned char)((iBufRef<<4)&0x00FF));
		        LCD_1_Position(0,8);
		        LCD_1_PrCString("P");
		        break;
		    case 4:
		        LCD_1_PrHexByte((unsigned char)(iBufPGain>>4));
		        break;
		    case 5:
		        LCD_1_PrHexByte((unsigned char)((iBufPGain<<4)&0x00FF));
		        LCD_1_Position(0,12);
		        LCD_1_PrCString("D");
		        break;
		    case 6:
		        LCD_1_PrHexByte((unsigned char)(iBufDGain>>4));
		        break;
		    case 7:
		        LCD_1_PrHexByte((unsigned char)((iBufDGain<<4)&0x00FF));
		        break;
		        
		    case 8:
		        LCD_1_Position(1,0);
		        LCD_1_PrHexInt(iAdValue);
		        break;
		    case 9:
		        LCD_1_PrCString("/");
		        LCD_1_PrHexInt(iPwmValue);
		        LCD_1_PrCString(" ");
		        break;
		    case 10:
		        switch(cCursor) {
		        case 0:
		        	LCD_1_PrCString("B");
		        	break;
		        case 1:
		        	LCD_1_PrCString("R");
		        	break;
		        case 2:
		        	LCD_1_PrCString("P");
		        	break;
		        case 3:
		        	LCD_1_PrCString("D");
		        	break;
		        default:
		        	break;
		        }
		        break;
		    case 11:
		        if(iPlusMinusFlag == 1)
		        	LCD_1_PrCString("+");
		        else
		        	LCD_1_PrCString("-");
			        LCD_1_PrHexByte(cBitHighLowFlag);
		        break;
		    default:
		        break;
	        }
        }else if(cModeFlag == 2) {
	        switch(cLCDFlag) {
	        case 0:
		        LCD_1_Position(0,0);
		        break;
		    case 1:
		        LCD_1_PrCString("Sla");
		        break;
		    case 2:
		        LCD_1_PrCString("ve ");
		        break;
		    case 3:
		        LCD_1_PrCString("Mod");
		        break;
		    case 4:
		        LCD_1_PrCString("e  ");
		        break;
		    case 5:
		        LCD_1_Position(1,0);
		        break;
		    case 6:
		        LCD_1_PrHexInt(iAdValue);
		        break;
		    case 7:
		        LCD_1_PrCString("/");
		        break;
		    case 8:
		        LCD_1_PrHexInt(iPwmValue);
		        break;
		    case 9:
		        LCD_1_PrCString("/");
		        break;
		    case 10:
		        LCD_1_PrHexInt(iAdValue2);
		        break;
		    case 11:
		        break;
		    default:
		        break;
	        }
        }
        if(++cLCDFlag==12) cLCDFlag=0;
    }
}
